
#include "RIF_2048_0,080_61.h"
//#define DEBUG // à décommenter pour afficher les échantillons et la TFD
#define FILTRAGE // à commenter pour enlever le filtrage
#define M 61 // nombre de coefficients du RIF
#define INTERPOL 1 // analyse des raies par interpolation quadratique
#define SEUIL_REL 0.08 // seuil relatif à l'harmonique la plus haute
#define P 11
#define Ne 2048 // 2^P nombre d'échantillons
#define PERIOD 100 // période d'échantillonnage en microsecondes
#define REDUC 5 // réduction de la fréquence d'échantillonnage après filtrage
#define VREF 3.3 // tension mesurée sur la sortie AREF

#define ADC_MR_TRIG1 (1<<1)
uint8_t nchannels;
#define MAX_NCHAN 1
uint8_t channels[MAX_NCHAN];

uint8_t gain[MAX_NCHAN];
uint8_t channels_pins[8] ={7,6,5,4,3,2,1,0};

const float duree = PERIOD*1e-6*Ne*REDUC; // durée du signal analysé en secondes


            
volatile float A_re[Ne];
float A_im[Ne];
float B_re[Ne];
float B_im[Ne];             
              
volatile float circulaire[M]; // liste circulaire pour le filtrage
uint16_t dernier; // indice du dernier échantillon dans la liste circulaire
uint8_t cr; // compteur pour la réduction
              
volatile uint16_t indice;
              
volatile char acquisition=0;
volatile char convol;
char start = 0;
uint8_t flag;
              
typedef struct raie {
  int indice; // indice dans le tableau TFD
  float pos; // indice réel extrapolé
  float amp; // amplitude
} Raie;              
              
void config_timer(uint32_t period) {
  uint32_t ticks = 42*period;
  uint32_t channel = 0;
  uint8_t clock = TC_CMR_TCCLKS_TIMER_CLOCK1; // horloge 84MHz/2=42 MHz
  pmc_enable_periph_clk (TC_INTERFACE_ID + 0*3+channel) ;
  TC_Configure(TC0,channel, TC_CMR_WAVE | TC_CMR_WAVSEL_UP_RC | TC_CMR_ACPA_CLEAR | TC_CMR_ACPC_SET |clock); 
  TC_SetRC(TC0,channel,ticks); 
  TC_SetRA(TC0,channel,ticks >> 1);
  TC_Start(TC0, channel);
}
              
void config_adc() {
    int k;
    pmc_enable_periph_clk(ID_ADC);
    adc_init(ADC, VARIANT_MCK, ADC_FREQ_MAX, 0);
    adc_configure_timing(ADC, 0, ADC_SETTLING_TIME_3, 1);
    adc_set_resolution(ADC,ADC_12_BITS);
    NVIC_EnableIRQ (ADC_IRQn) ;   
    ADC->ADC_IDR=0xFFFFFFFF ;  
    ADC->ADC_CHDR = 0xFFFF; 
    ADC->ADC_CHSR = 0x0000; 
    ADC->ADC_WPMR &= ~0x1;
    ADC->ADC_CHER = 0x0000;
    ADC->ADC_SEQR1 = 0x00000000;
    ADC->ADC_CGR = 0x00000000;
    ADC->ADC_COR = 0x00000000;
    for (k=0; k<nchannels; k++) {
      ADC->ADC_SEQR1 |= ((channels[k]&0xF)<<(4*k));
      ADC->ADC_CHER |= (1<<k);
      if (gain[k]==2) ADC->ADC_CGR |= 2 << (2*k);
      if (gain[k]==4) ADC->ADC_CGR |= 3 << (2*k);
      ADC->ADC_COR |= 1 << k; 
    }
    ADC->ADC_MR = (ADC->ADC_MR & 0xFFFFFFF0) | ADC_MR_TRIG1 | ADC_MR_TRGEN | ADC_MR_USEQ;
    ADC->ADC_IDR=~(1<<channels[0]); 
    ADC->ADC_IER=1<<channels[0];
}
              
void ADC_Handler() {
  circulaire[dernier] = *(ADC->ADC_CDR+channels[0]);
    cr += 1;
  if (convol&&(cr==REDUC)) {
      cr = 0;
      #ifdef FILTRAGE
      int16_t k = dernier;
      int16_t m = 0;
      float accum = 0;
      while (m<M) {
          accum += circulaire[k]*rif[m];
          k -= 1;
          if (k==-1) k=M-1;
          m += 1;
      }
      A_re[indice] = accum;
      #else
      A_re[indice] = circulaire[dernier];
      #endif
      indice += 1;
      if (indice==Ne) {
          TC_Stop(TC0,0);
          pmc_disable_periph_clk (TC_INTERFACE_ID);
          acquisition = 0;
          
      }
  }
  if (cr==M) {convol=1; cr=0;}
  dernier += 1;
  if (dernier==M) {
      dernier=0;
  }
}           
               
int inversion(uint16_t i) {
  uint8_t bi;
  int8_t b;
  uint8_t bits[P];
  for (b=0; b<P; b++) {
    bi = i & 0b1;
    i = i >> 1;
    bits[b] = bi;
  }
  uint16_t facteur = 1;
  uint16_t j = 0;
  for (b=P-1; b>=0; b--) {
    j += bits[b]*facteur;
    facteur *= 2;
  }
  return j;
}         
                  
void fft() {
  uint16_t j;
  for (int k=0; k<Ne; k++) {
      j = inversion(k);
      B_re[j] = A_re[k];
      B_im[k] = A_im[k];
  }
  uint16_t taille = 1;
  uint16_t taille_precedente;
  uint16_t nombre_tfd = Ne;
  uint16_t position;
  float phi, W_re, W_im;;
  float x,y;
  uint8_t parite = 0;
  for (uint16_t q=0; q<P; q++) {
      taille_precedente = taille;
      taille *= 2;
      nombre_tfd /= 2;
      for (uint16_t m=0; m<nombre_tfd; m++) {
          phi = -2*PI/taille;
          position = m*taille;
          uint16_t i,j,k;
          for (i=0; i<taille_precedente; i++) {
              W_re = cos(phi*i);
              W_im = sin(phi*i);
              j = position+taille_precedente+i;
              k = position +i;
              if (parite==0) {
                x = B_re[j];
                y = B_im[j];
                A_re[k] = B_re[k] + x*W_re-y*W_im;
                A_im[k] = B_im[k] + x*W_im+y*W_re;
              }
              else {
                x = A_re[j];
                y = A_im[j];
                B_re[k] = A_re[k] + x*W_re-y*W_im;
                B_im[k] = A_im[k] + x*W_im+y*W_re;
              }
          }
          for (i=taille_precedente; i<taille; i++) {
              W_re = cos(phi*i);
              W_im = sin(phi*i);
              j = position+i;
              k = position+i-taille_precedente;
              if (parite==0) {
                x = B_re[j];
                y = B_im[j];
                A_re[j] = B_re[k] + x*W_re-y*W_im;
                A_im[j] = B_im[k] + x*W_im+y*W_re;
              }
              else {
                x = A_re[j];
                y = A_im[j];
                B_re[j] = A_re[k] + x*W_re-y*W_im;
                B_im[j] = A_im[k] + x*W_im+y*W_re;
              }   
          }  
      }
      parite = ~parite;
  }
  if (parite) {
      for (int k=0; k<Ne; k++) {B_re[k] = A_re[k]; B_im[k] = A_im[k];}
  }
}                 
                  
void recherche_max(Raie *raie, float seuil, bool interpol) {
     uint16_t imax = 1;
     float module;
     float module_max = 0;
     for (uint16_t k=1; k<Ne/2; k++) {
        module = B_re[k];
        if (module > module_max) {
          imax = k;
          module_max = module;
        }
     }
     if (module_max < seuil) {raie->indice = -1; return;}
     if (interpol) {
       float a,b,c;
       a = B_re[imax-1];
       b = module_max;
       c = B_re[imax+1];
       if ((a>seuil)&&(c>seuil)) {
         raie->indice = imax;
         float p = 0.5*(a-c)/(a-2*b+c);
         raie->amp = b-0.25*(a-c)*p;
         raie->pos = imax+p;
         return;
       }
     }
     
     raie->indice = imax;
     raie->pos = imax;
     raie->amp = module_max;
     
}                  
                  
void effacer_raie(int imax, float seuil) {
    int i = imax;
    while ((i>0)&&(B_re[i]>seuil)) {
        B_re[i] = 0;
        i -= 1;
    }
    i = imax + 1;
    while ((i<Ne/2)&&(B_re[i]>seuil)) {
        B_re[i] = 0;
        i += 1;
    }
}                   
                   
void setup() {
  Serial.begin(115200);
  start = 0;
  
}                   
                   
void loop() {
  delay(100);
  if (start==0) {    
        for (uint32_t i=0; i<Ne; i++) {A_re[i] = 0; A_im[i] = 0;}
        start = 1;
        flag = 0;
        acquisition=1;
        indice = 0;
        dernier = 0;
        cr = 0;
        convol = 0;
        nchannels = 1;
        channels[0] = channels_pins[0];
        gain[0] = 0;
        config_adc();
        config_timer(PERIOD);
   }
  if (acquisition==0) {
     #ifdef DEBUG
     Serial.println("----------------------");
     for (uint16_t k=0; k<Ne; k++) {
          Serial.print((float)k*PERIOD*REDUC*1e-6,4);Serial.print(" "); Serial.println(A_re[k]*VREF/4096);
     }
     #endif
     fft();
     for (uint16_t k=0; k<Ne/2; k++) {
       #ifdef FILTRAGE 
       B_re[k] = sqrt(B_re[k]*B_re[k]+B_im[k]*B_im[k])*2.0/Ne/gain_filtre[k];
      #else
       B_re[k] = sqrt(B_re[k]*B_re[k]+B_im[k]*B_im[k])*2.0/Ne;
      #endif
     }
     #ifdef DEBUG
     Serial.println("----------------------");
     for (uint16_t k=0; k<Ne/2; k++) {
         Serial.print(k/duree,4);Serial.print(" "); Serial.println(B_re[k]*VREF/4096,4);
     }
     #endif
     float seuil = 20; // seuil initial
     char premier = 1;
     Raie raie;
     float f;
     float amp_max;
     Serial.println("-------------------");
     do {
        recherche_max(&raie,seuil,INTERPOL);
        if (premier) {
            premier = 0;
            seuil = raie.amp * SEUIL_REL; // réglage du seuil par rapport à l'harmonique de plus grande amplitude
            amp_max = raie.amp;
        }
        if (raie.indice>0) {
            effacer_raie(raie.indice,seuil);
            f = raie.pos*1/duree;
            if (raie.amp>0)
              {
                Serial.print("f = "); Serial.print(f); Serial.print(" +/-"); Serial.print(0.5/duree);
                Serial.print(",  A = "); Serial.print(raie.amp*VREF/4096);
                Serial.print("  ("); Serial.print(20*log10(raie.amp/amp_max)); Serial.println(" dB)");
              }
        }
     } 
     while (raie.indice>0);
     start = 0;
  }
}
                   